#!/usr/bin/perl -w

use strict;

use DBI;
use Digest::MD5 qw(md5_hex);
use File::Path qw(remove_tree);
require LWP::UserAgent;
use POSIX qw(uname);

my $arch = (uname())[4];

my $tmpdir = "/tmp/bluesnap" . int(rand(9999));

my $dbdir = "/var/db/bluesnapper";
my $webhost = "http://snapsnap.tedunangst.com/$arch";

my $dbh;
my $diffstmt;
my $dirstmt;
my $filestmt;
my $linkstmt;
my @mismatches;

sub init {
	my $dbfile = shift;
	die "where's the db?" unless -f $dbfile;
	$dbh = DBI->connect("dbi:SQLite:$dbfile") or die "no db";
	$diffstmt = $dbh->prepare("select file, diff, md5 from diffs");
	$dirstmt = $dbh->prepare("select dir, uid, gid, mode from dirs");
	$filestmt = $dbh->prepare("select file, data, uid, gid, mode " .
		"from files");
	$linkstmt = $dbh->prepare("select src, dst, sym from links");
}

sub uniq {
	my %seen;
	my @out;
	while (my $el = shift) {
		push @out, $el if !$seen{$el};
		$seen{$el} = 1;
	}
	return @out;
}

sub writeall {
	my $file = shift;
	my $data = shift;

	my $fd;
	open $fd, ">$file" or die "can't open $file";
	syswrite $fd, $data;
	close $fd;
}

sub readall {
	my $file = shift;

	my @info = stat $file or die "where'd $file go?";
	my $size = $info[7];
	my $fd;
	open $fd, $file;

	my $data;
	read $fd, $data, $size;
	close $fd;
	return $data;
}

sub md5file {
	my $file = shift;
	return md5_hex(readall $file);
}

sub setinfo {
	my $file = shift;
	my $uid = shift;
	my $gid = shift;
	my $mode = shift;

	chown $uid, $gid, $file;
	chmod $mode, $file;
}

sub patch {
	my $ofile = shift;
	my $patch = shift;
	my $csum = shift;

	if (! -f $ofile || $csum ne md5file($ofile)) {
		if (-f "/bsd.mp" && $ofile eq "/bsd" && $csum eq md5file("/bsd.mp")) {
			$ofile = "/bsd.mp";
		} elsif (-f "/bsd" && $ofile eq "/bsd.mp" && $csum eq md5file("/bsd")) {
			$ofile = "/bsd";
		} else {
			push @mismatches, $ofile;
			print "md5 mismatch on $ofile.  Skipping.\n";
			return;
		}
	}
	my $nfile = $ofile . ".newsnap";

	print "patching $ofile\n";
	unlink $nfile;
	my $pfile = "$tmpdir/bspfile";
	writeall $pfile, $patch;

	if ($ofile =~ /.*\.gz$/) {
		system "gzcat $ofile > $tmpdir/ungzo";
		system "bspatch $tmpdir/ungzo $tmpdir/ungzn $pfile";
		system "gzip -o $nfile $tmpdir/ungzn";
	} else {
		system "bspatch $ofile $nfile $pfile";
	}
	my @info = stat $ofile;

	setinfo $nfile, $info[4], $info[5], $info[2];

	rename $nfile, $ofile;
}

sub create {
	my $ofile = shift;
	my $data = shift;
	my $uid = shift;
	my $gid = shift;
	my $mode = shift;
	my $nfile = $ofile . ".newsnap";

	if (-e $ofile) {
		print "$ofile already exists.  Not creating.\n";
		return;
	}
	print "creating $ofile\n";

	writeall $nfile, $data;
	setinfo $nfile, $uid, $gid, $mode;
	rename $nfile, $ofile;
}

sub makelink {
	my $src = shift;
	my $targ = shift;
	my $sym = shift;
	my $nfile = $src . ".newsnap";

	unlink $nfile;

	if ($sym) {
		symlink $targ, $nfile;
	} else {
		link $targ, $nfile;
	}

	rename $nfile, $src;
	unlink $nfile; # rename leaves nfile behind if same file
}

sub createdir {
	my $dir = shift;
	my $uid = shift;
	my $gid = shift;
	my $mode = shift;
	my $nfile = $dir . ".newsnap";

	if (-d $dir) {
		print "$dir already exists.  Not creating.\n";
		return;
	}
	mkdir $nfile;
	setinfo $nfile, $uid, $gid, $mode;
	rename $nfile, $dir;
}

sub dopatching {

	$dirstmt->execute;
	while (my $row = $dirstmt->fetchrow_arrayref) {
		createdir($$row[0], $$row[1], $$row[2], $$row[3]);
	}
	$dirstmt->finish;

	$diffstmt->execute;
	while (my $row = $diffstmt->fetchrow_arrayref) {
		my $ofile = $$row[0];
		my $data = $$row[1];
		my $m = $$row[2];
		patch($ofile, $data, $m);
	}
	$diffstmt->finish;

	$filestmt->execute;
	while (my $row = $filestmt->fetchrow_arrayref) {
		my $ofile = $$row[0];
		my $data = $$row[1];
		my $uid = $$row[2];
		my $gid = $$row[3];
		my $mode = $$row[4];
		create($ofile, $data, $uid, $gid, $mode);
	}
	$filestmt->finish;

	$linkstmt->execute;
	while (my $row = $linkstmt->fetchrow_arrayref) {
		my $src = $$row[0];
		my $dst = $$row[1];
		my $sym = $$row[2];
		makelink($src, $dst, $sym);
	}
	$linkstmt->finish;

	$dbh->disconnect;
}

sub httpget {
	my $url = shift;
	my $ua = LWP::UserAgent->new;
	$ua->default_header('Accept-Encoding' =>
		scalar HTTP::Message::decodable());
	my $resp = $ua->get($url);
	die "request to $url failed" unless $resp->is_success;
	return $resp->decoded_content;
}

sub determinesnap {
	my $latestfile = "$dbdir/latest.txt";
	my $snapnum = 0;
	if (-e $latestfile) {
		print "latest.txt found, using that.\n";
		$snapnum = int(readall $latestfile);
	} else {
		print "Attempting to identify snap.\n";
		my $fingers = httpget "$webhost/snap-fingers.txt";
		for my $line (split '\n', $fingers) {
			my ($file, $md5, $num) = split ' ', $line;
			if (-f $file && $md5 eq md5file($file)) {
				print "Found snap $num.\n";
				$snapnum = $num;
				last;
			}
		}
	}
	print "Currently at snap $snapnum.\n";
	return $snapnum;
}

sub findpatches {
	my $snapnum = determinesnap;

	die "unknown snapnum" if ! $snapnum;

	my $lastsnap = int(httpget "$webhost/latest.txt");
	if ($lastsnap == $snapnum) {
		print "Already at latest.\n";
		return;
	}
	print "Will upgrade to $lastsnap.\n";
	while ($snapnum < $lastsnap) {
		$snapnum++;
		my $dbfile = "diff$snapnum.db";
		print "downloading snap $snapnum...\n";
		my $dbdata = httpget "$webhost/$dbfile";
		writeall "$tmpdir/$dbfile", $dbdata;
		print "patching...\n";

		init "$tmpdir/$dbfile";
		dopatching;
	}

	chdir "$dbdir/etc";
	system "tar cfz $tmpdir/etc.tgz ./*";
	chdir "$dbdir/xetc";
	system "tar cfz $tmpdir/xetc.tgz ./*";
	system "sysmerge -s $tmpdir/etc.tgz -x $tmpdir/xetc.tgz";

	if (@mismatches) {
		print "The following files were not updated.\n";
		for my $mm (uniq @mismatches) {
			print "$mm\n";
		}
	}

	print "All done.\n";
	writeall "$dbdir/latest.txt", $snapnum;

}

sub main {

	mkdir $tmpdir, 0700 or die "can't make tmpdir $tmpdir";

	findpatches;

	remove_tree($tmpdir);

}

main;
